@@ -15,7 +15,7 @@ class VectorBrusselator(Vector):
     """
 
     def __init__(self):
-        super(VectorBrusselator, self).__init__()
+        super().__init__()
         self.value = np.zeros(2)
 
     def set_values(self, value):
@@ -24,6 +24,11 @@ class VectorBrusselator(Vector):
     def get_values(self):
         return self.value
 
+    def clone(self):
+        tmp = VectorBrusselator()
+        tmp.set_values(self.get_values())
+        return tmp
+
     def clone_zero(self):
         return VectorBrusselator()
 
@@ -42,9 +47,20 @@ class VectorBrusselator(Vector):
         tmp.set_values(self.get_values() - other.get_values())
         return tmp
 
+    def __mul__(self, other):
+        tmp = VectorBrusselator()
+        tmp.set_values(self.get_values() * other)
+        return tmp
+
     def norm(self):
         return np.linalg.norm(self.value)
 
+    def pack(self):
+        return self.value
+
+    def unpack(self, value):
+        self.value = value
+
     def plot_solution(self):
         plt.plot(self.value[0], self.value[1], color='red', marker='o')
 
@@ -75,7 +91,7 @@ class Brusselator(Application):
     """
 
     def __init__(self, *args, **kwargs):
-        super(Brusselator, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
         self.vector_template = VectorBrusselator()
         self.vector_t_start = VectorBrusselator()
 
